#include "doomdef.h"
#include "g_game.h"
#include "p_local.h"
#include "p_inter.h"
#include "s_sound.h"
#include "m_random.h"
#include "w_main.h"   //faB: in dev.
#include "r_main.h"


//Init Monster data.
char *monstername[NUM_SPAWNTYPES] = {"barricade", "possessed","seargant", "chaingunner", "imp",	"deamon",	"spector", "cacodeamon", "painelemental", "hellknight", "hellbaron", "revenant", "mancubus", "arachnotron",	"mastermind",	"archvile", "cyberdeamon"};
mobjtype_t monsters[NUM_SPAWNTYPES] = {MT_BBARREL,MT_POSSESSED,MT_SHOTGUY,MT_CHAINGUY,MT_TROOP,MT_SERGEANT,MT_SHADOWS,MT_HEAD,MT_PAIN,MT_KNIGHT,MT_BRUISER,MT_UNDEAD,MT_FATSO,MT_BABY,MT_SPIDER,MT_VILE,MT_CYBORG};
int monsterprice[NUM_SPAWNTYPES] = {10 ,1, 3, 50, 5, 6, 8, 35, 350, 70, 90, 85, 75, 125, 550, 900, 700};

short localmousex = 0;
short localmousey = 0;


boolean Spawn_Monster(int type, player_t *player, angle_t angle)
{
    fixed_t     x;
    fixed_t     y;
    fixed_t     z;

	mobj_t*		actor;
    mobj_t*     newmobj;
    angle_t     an;
    int         prestep;

	if(!player)
    	I_Error("Tech priest was not a player\n");

	if(player->resources<monsterprice[type]) //Cant afford.
		return false;

	actor = player->mo;

    // okay, there's place for another one
    an = angle >> ANGLETOFINESHIFT;

    prestep =
        4*FRACUNIT
        + 3*(actor->info->radius + mobjinfo[monsters[type]].radius)/2;

    x = actor->x + FixedMul (prestep, finecosine[an]);
    y = actor->y + FixedMul (prestep, finesine[an]);
    z = actor->z + 8*FRACUNIT;

    newmobj = P_SpawnMobj (x , y, z, monsters[type]);

	newmobj->angle = an;
    newmobj->target = NULL;
	newmobj->owner  = actor;
    // kill it immediately
    if (!P_TryMove (newmobj, newmobj->x, newmobj->y, false))
    {
        P_RemoveMobj (newmobj);
        return false;
    }

    newmobj->flags |= (player->skincolor)<<MF_TRANSSHIFT;

	//Pay for it.
	player->resources -= monsterprice[type];

	return true;
}



void SetMovetarget(mobj_t *actor, fixed_t x, fixed_t y)
{
	if(!actor->movetarget) 
		actor->movetarget = P_SpawnMobj(x, y, ONFLOORZ, MT_NODE);
	else
	{
		P_UnsetThingPosition (actor->movetarget);
		actor->movetarget->x=x,
		actor->movetarget->y=y,
		actor->movetarget->z=ONFLOORZ,
		P_SetThingPosition (actor->movetarget);
	}
	actor->moving = true;
	actor->guarding = false;
	actor->follows = false;
	P_SetMobjState(actor, actor->info->seestate);
}


void SetEnemy(mobj_t *actor, mobj_t *target)
{
	if(actor->type==MT_BBARREL)
		return;
	actor->target = target;
	if(actor->target)
		P_SetMobjState(actor, actor->info->seestate);
		
}


void DoMana(mobj_t *mo)
{
	int i, mana;
	
	for(i=0; i<MAXPLAYERS; i++)
	{
		if(!playeringame[i] || !players[i].mo)
			continue;		

		if(players[i].resources>=MAXMANA) 
			return;
		if(P_AproxDistance(mo->x-players[i].mo->x, mo->y-players[i].mo->y) < GETDISTANCE) //Close enough to suck mana.
		{
			mana = MANAPOINT;
			players[i].resources+=mana;
			if(players[i].resources>MAXMANA) 
				players[i].resources = MAXMANA;
			mo->health-=mana;
			if(mo->health<=0) //Mana exhausted. 
			{
				P_RemoveMobj(mo);
				return;
			}
		}
	}
}


boolean Check_LOS (mobj_t *from, mobj_t *to, angle_t vangle)
{
	if (!P_CheckSight (from, to))
		return false; // out of sight
	if (vangle == ANGLE_MAX)
		return true;
	if (vangle == 0)
		return false; //Looker seems to be blind.

	return abs (R_PointToAngle2 (from->x, from->y, to->x, to->y) - from->angle) <= (int)vangle/2;
}


//Removes all the monsters that belonged to the owner whois recently executed.
void PlayerDeath(mobj_t *owner)
{
    thinker_t *th;
	mobj_t *mo;
	if(!owner) return;
	for (th = thinkercap.next ; th != &thinkercap ; th=th->next)
    {
		mo = (mobj_t *)th;

		if(IsMonster(mo))
			mo->target = NULL; //Monsters should all get new targets, since many will be invalid.
		
		if(owner == mo->owner)
			P_RemoveMobj(mo);		
	}
}

mobj_t *looker = NULL;
void MonsterExtraThink(mobj_t *count)
{
//	static int c;
	
	if(count->type==MT_MANA && count->tics == 1)
	{
		DoMana(count);
		return;
	}		

	if(!IsMonster(count))
		return;
		
	if(looker && looker->health<=0)
		looker = NULL;

	if(!looker && count->health>0) //Assign new looker.
	{
		looker = count;
//		CONS_Printf("%i\n", c);
//		c = 0;
		return;
	}
	if(count==looker) //Gone a whole loop.
	{
		looker=NULL;
		return;
	}

	if(!looker)
		return;

//	c++;		

	//Special things for special monsters.
	if(looker->type == MT_VILE)
	{ //lets vile roam towards dead friends, inorder to raise them.
		if(!looker->moving)
			if(count->owner==looker->owner && count->health<=0 && count->info->raisestate!=S_NULL)
				SetMovetarget(looker, count->x, count->y);
	}
	else if(looker->type==MT_SHADOWS || looker->type==MT_SERGEANT)
		looker->guarding = false;
	/////////////


	if(count->health<=0) //From noone, dead counts are not counted.)
		return;
	
	//trash invalid (dead or owner)  enemies.
	if(looker->target && looker->target->health<=0)
   		looker->target = NULL;
	if(looker->target == looker->owner)
   		looker->target = NULL;

	//Need a new enemy?
	if(count->owner!=looker->owner)
	{
		if( (looker->target && !P_CheckSight(looker, looker->target) && P_CheckSight(looker, count)) || !looker->target)
			SetEnemy(looker, count);
	}

	//give target to friends
	if(count->owner==looker->owner && looker->target && !count->target)
		SetEnemy(count, looker->target);

	//Movetarget handling.
	if(looker->movetarget && looker->moving)
	{
		looker->target = looker->movetarget;
		if(P_AproxDistance(looker->x-looker->movetarget->x, looker->y-looker->movetarget->y) < 128*FRACUNIT)
		{
			looker->guarding = true;
			looker->moving = false;
			looker->target = NULL;
		}
	}	
}


void SetFollow(mobj_t *actor)
{
    thinker_t *th;
	mobj_t *mo;

	for (th = thinkercap.next ; th != &thinkercap ; th=th->next)
    {
		mo = (mobj_t *)th;
		
		if(mo->owner!=actor)
			continue;
		if(!IsMonster(mo))
			continue;
		if(mo->health<=0)
			continue;
		if(mo->type==MT_BBARREL)
			continue;

		if(!Check_LOS(actor, mo, ORDERANGLE))
			continue;

		mo->follows = true;
		mo->guarding = false;
		mo->moving = false;
	}
}

void SetGuard(mobj_t *actor)
{
    thinker_t *th;
	mobj_t *mo;

	for (th = thinkercap.next ; th != &thinkercap ; th=th->next)
    {
		mo = (mobj_t *)th;
		
		if(mo->owner!=actor)
			continue;
		if(!IsMonster(mo))
			continue;
		if(mo->health<=0)
			continue;
		if(mo->type==MT_BBARREL)
			continue;
		if(!Check_LOS(actor, mo, ORDERANGLE))
			continue;

		mo->moving = false;
		mo->guarding = !mo->guarding;
		if(mo->guarding) mo->follows = false;
		P_SetMobjState(mo, mo->info->painstate);
	}
}

void W_Deselect(player_t *player)
{
    thinker_t*          th;
	mobj_t *mo;

	for (th = thinkercap.next ; th != &thinkercap ; th=th->next)
    {
		mo = (mobj_t *)th;

		if(!mo->owner) 
			continue;
		
		if(mo->owner==player->mo)
			mo->selected = false;
	}
}	


//Select monsters within select box
void W_DoSelect(player_t *player)
{
    thinker_t*          th;
	mobj_t *mo;

	W_Deselect(player);

	for (th = thinkercap.next ; th != &thinkercap ; th=th->next)
    {
		mo = (mobj_t *)th;

		if(!mo->owner) 
			continue;
		
		if(mo->health<=0)
			continue;

		if(mo->owner==player->mo && mo->type!=MT_BBARREL)
		{			
			if(mo->x>player->sx && mo->x<player->mx && mo->y<player->sy && mo->y>player->my)
				mo->selected = true;
		}
	}
}	


void W_OrderMove(player_t *player)
{
    thinker_t*          th;
	mobj_t *mo;

	for (th = thinkercap.next ; th != &thinkercap ; th=th->next)
    {
		mo = (mobj_t *)th;

		if(mo->owner==player->mo && mo->type!=MT_BBARREL && mo->selected)
			SetMovetarget(mo, player->mx, player->my);
	}
}

void W_MapControl(player_t *player)
{
	boolean lb;
	
	if(!(player->cmd.buttons2 & BT_MAP))
		return;

	lb = (player->cmd.buttons & BT_ATTACK) ? true : false;

	player->mx += player->cmd.mousex*FRACUNIT;
	player->my += player->cmd.mousey*FRACUNIT;

	if(!player->hold)
	{
		player->sx = player->mx;
		player->sy = player->my;
	}
		
	if(lb) player->hold = true;	

	if(player->hold && !lb)	 //On relase
	{
		if(abs(player->sx-player->mx)>FRACUNIT*3 && abs(player->sy-player->my)>FRACUNIT*3) //Do select 
			W_DoSelect(player);
		else
			W_OrderMove(player);

		player->hold = false;		
	}
}








